c     -*- mode: FORTRAN -*-
c
c     This file is part of krot,
c     a program for the simulation, assignment and fit of HRLIF spectra.
c
c     Copyright (C) 1994-1998 Arnim Westphal
c     Copyright (C) 1997,1998 Jochen Kpper
c     
c
c     krot-arnirot
c     a program to calculate rotational resolved vibrational/vibronic bands
c
c     this program utilizes the Watson-Hamiltonian up to quartic terms
c     for reference see: James K. G. Watson, Vibrational Spectra and Structure 6, 1 (1977), ed J.R. Durig)


#include <arni.h>


c     (main) routine to calculate two state rovibronic spectra
      subroutine arnical( Jmx, dmevl, dmevc, maxnl, npr, ntheli, icqn,
     *                    swg, swe, swang, itrvec,
     *                    diaalg,
     *                    evalg, evale,
     *                    evecgr, evecgi, evecer, evecei,
     *                    Jmxcal, dKmax, cutint,
     *                    lifile, asfile,
     *                    fitges, ifit, sigma,
     *                    polori,
     *                    shorti,
     *                    cRotCg, cRotCe, cDRotC, cState, cEuler, no_yes, off_on,
     *                    frame, kind,
     *                    cBran, polax1, polax2,
     *                    evnodd, symdes,
     *                    lngbar, lstars, lbars, ldash,
     *                    lqn, lqn2, intens, thefre, expfre,
     *                    temp1, temp2, weight, nuspsw,
     *                    nuzero,
     *                    divis, Jmxout, normf,
     *                    rotcog, rotcoe, idelta,
     *                    ivpt )

      implicit none

c     maximum (possible) rotational quantum number
      integer        Jmax, Jmx
      parameter      ( Jmax = ARNIROT_JMAX )

c     maximum dimension of hamiltonian matrix
      integer        dmham
      parameter      ( dmham = 2*Jmax + 1 )
c     maximum dimension of factorized Wang hamiltonian submatrix
      integer        dmfwng
      parameter      ( dmfwng = Jmax/2 + 1 )
c     dimension of eigenvalue storage vectors
      integer        dmeval, dmevl
      parameter      ( dmeval = (Jmax + 1)*(Jmax + 1) )

c     number of elements for eigenvector storage
      integer        dmevec, dmevc
c     help variable for the calculation of dmevec
      real*8         szvec
c     last summand is 1.5 instead of 1.0 to prevent roundoff error
      parameter      ( szvec = ((4.d0/3.d0*Jmax + 4.d0)*Jmax + 11.d0/3.d0)*Jmax + 1.5d0 )
      parameter      ( dmevec = szvec )

c     maximum number of lines
      integer        maxnli, maxnl
      parameter      ( maxnli = ARNIROT_MAXNLI )
c     maximum number of rotational model parameters (npar each state plus origin)
      integer        npar, npr
      parameter      ( npar = ARNIROT_NPAR )


c     switch for diagonalization algorithm
      integer        diaalg
c     switch for fitting the parameters of ground and/or excited state
      integer        fitges
c     loop variables
      integer        i, k
c     actual size of J block (2J+1)
      integer        iact
c     actual size of greatest submatrix (J/2 + 1)
      integer        iactr
c     flags for complex hamiltonians
      integer        icome, icomg
c     counter for varied parameters
      integer        icount
c     rotational state kets |J Ka Kc>
      integer        icqn(dmevl,3)
c     flag for delta rotational constants
      integer        idelta
c     pointer on first eigenvalue of a given J as contained in eval?
      integer        iept
c     individual fitting switches for each rotational model parameter
      integer        ifit(2*npr+1)
      integer        ifitpt(2*npar+1)
c     ordering pointer for sorting
      integer        iop(maxnli)
c     index to theoretical lines with experimental assignments (TheExpMatch)
      integer        ipttem(maxnli)
c     sign of transition moment component
      integer        isign
c     transform transition moment flag
      integer        itrvec
c     pointer on the first element of the eigenvector block of a given J as contained in evec??
      integer        ivpt(0:Jmx)
c     current rotational quantum number
      integer        J
c     Jmax for calculation
      integer        Jmxcal
c     Jmax for output of matrices
      integer        Jmxout
c     Delta K max
      integer        dKmax
c     warning flag for overflow of lines data
      integer        lioflw
c     set of quantum numbers for a transition
c     lqn  : J', Ka', Kc', J", Ka", Kc"
c     lqn2 : type (1=a, 2=b, 3=c), Ka", Delta Ka, J"-Kc", Delta J, Kc" (old)
c     lqn2 : type (1=a, 2=b, 3=c), Delta J, Delta Ka, Ka", J"-Kc", Kc" (new)
      integer        lqn(maxnl,6), lqn2(maxnl,6)
c     number of matching data sets (transitions)
      integer        ndata
c     number of assigned lines
      integer        nexpli
c     normalization factor for line intensities
      integer        normf
c     number of actually used rotational model parameters (<= npar)
      integer        npare, nparg
c     number of parameters to be actually fit (<= 2npar+1)
      integer        nparft
c     number of calculated lines
      integer        ntheli
c     nuclear spin statistical weights  ee eo oe oo
      integer        nuspsw(0:3)
c     flag for new IO format
      integer        shorti
c     flags for axis reorientation
      integer        swg, swe


c     SVD solution vector x
      real*8         a(2*npar+1), atmp(2*npar+1)
c     submatrices E-, E+, O-, O+
      real*8         aem(dmfwng,dmfwng), aep(dmfwng,dmfwng)
      real*8         aom(dmfwng,dmfwng), aop(dmfwng,dmfwng)
c     chi square
      real*8         chisq
c     intensity cut off
      real*8         cutint, cutor
c     covariance matrix
      real*8         cvm(2*npar+1,2*npar+1)
c     eigenvalues of submatrices E-, E+, O-, O+
      real*8         dem(dmfwng), dep(dmfwng), dom(dmfwng), dop(dmfwng)
c     derivatives d(eval?)/d(rotco?) corresponding to each eigenvalue
      real*8         derivg(dmeval,npar), derive(dmeval,npar)
c     derivative information assigned to calculated transition
      real*8         derlg(maxnli,npar), derle(maxnli,npar)
c     divisor for line frequencies
      real*8         divis
c     vector space for temporary eigenvalue storage
      real*8         e(dmham)
      real*8         er(dmfwng)
      real*8         eb(2*dmham)
c     eigenvalues of complex hamiltonian matrix (complete / real / imaginary)
      real*8         eneb(2*dmham)
      real*8         ener(dmham), enei(dmham)
#ifdef DEBUG_VERBOSE
c     eigenvalues calculated using derivatives
      real*8         ened(dmham)
#endif
c     eigenvalues (final storage)
      real*8         evalg(dmevl),  evale(dmevl)
c     eigenvectors (final storage)
c     - real part
      real*8         evecgr(dmevc), evecer(dmevc)
c     - imaginary part
      real*8         evecgi(dmevc), evecei(dmevc)
c     uncertainties of fitted parameters
      real*8         err(2*npar+1)
c     observed line positions
      real*8         expfre(maxnl)
c     summarized derivatives: -dE''/d lambda''   (1...npar)
c                              dE' /d lambda'    (npar+1...2*npar)
c                              ddE /d nuzero = 1 (2*npar+1)
      real*8         funcs(maxnli,2*npar+1), funcst(maxnli,2*npar+1)
c     complex hamiltonian matrix (complete / real part / imaginary part)
      real*8         hamb(2*dmham,2*dmham)
      real*8         hamr(dmham,dmham), hami(dmham,dmham)
c     line intensities
      real*8         intens(maxnl)
c     band origin
      real*8         nuzero
c     transition dipole components
      real*8         polori(3), polcal(3)
c     packed quantum numbers
      real*8         pqnth(maxnli)
c     rotational constants
      real*8         rotcog(npr), rotcoe(npr)
c     axis switched rotational constants tensor
      real*8         rraxsw(6,2)
c     sum of squared deviations in line positions (obs - calc)
      real*8         sdsum
c     common standard deviation for all experimental lines
      real*8         sigma
c     axis SWitching Euler ANGles theta,phi,chi g.s./e.s.
      real*8         swang(3,2)
c     temperatures and weighting factor
      real*8         temp1, weight, temp2
c     calculated line positions
      real*8         thefre(maxnl)
c     workspace provided for subroutine chevec
c     real*8         tmpvec(dmham)
c     F to g rotation and inverse rotation matrices (ZARE eq.(3.36))
      real*8         trang(3,3), trane(3,3), tring(3,3), trine(3,3)
c     SVD matrices U and V
      real*8         u(maxnli,2*npar+1)
      real*8         v(2*npar+1,2*npar+1)
c     matrices for eigenvectors (total complex / real part / imaginary part)
      real*8         vecb(2*dmham,2*dmham)
      real*8         vecr(dmham,dmham), veci(dmham,dmham)
c     eigenvectors of submatrices E-, E+, O-, O+
      real*8         vem(dmfwng,dmfwng), vep(dmfwng,dmfwng)
      real*8         vom(dmfwng,dmfwng), vop(dmfwng,dmfwng)
c     diagonal elements of SVD matrix W
      real*8         w(2*npar+1)
c     hamiltonian matrix in Wang basis (real / imaginary)
      real*8         wanghr(dmham,dmham), wanghi(dmham,dmham)
c     Wang transformation matrix
      real*8         wangtr(dmham,dmham)
c     right hand side vector b of linear equations system
      real*8         y(maxnli)


c     branch / parity / axes / symmetry species / logic etc. symbols
c     cf. subroutine setup
      character*1    cBran(-1:1), evnodd(0:1), polax1(3), polax2(3)
      character*2    symdes(0:1,0:1)
      character*3    no_yes(0:1), off_on(0:1)
      character*5    cEuler(3)
      character*7    cState(2)
      character*8    cRotCg(npar), cRotCe(npar)
      character*9    kind(0:1)
      character*10   frame(0:1)
      character*12   cDRotC(npar)
      character*81   lstars, lbars, ldash
      character*220  lngbar
c     file names
      character*250  asfile, lifile

c     common blocks
c     axis switching Euler angles and flags
c     common/ axswi / swg, swe, swang, itrvec
c     energy derivative information
c     common/ derivs/ derivg, derive, derlg, derle
c     ground and excited state energy values (eigenvalues)
c     common/ evals / evalg, evale
c     ground and excited state eigenvector coefficients
c     common/ evecs / evecgr, evecgi, evecer, evecei
c     extent of calculation
c     common/ extcal/ Jmxcal, dKmax, cutint
c     file names
c     common/ files / lifile, asfile
c     fit parameters (switches)
c     common/ fit   / fitges, ifit, sigma
c     ground state polarization percentages
c     common/ hybrid/ polori, polcal
c     flag for new IO format
c     common/ IOform/ shorti
c     frequently used labels etc.
c     common/ label1/ cRotCg, cRotCe, cDRotC, cState, cEuler, no_yes, off_on
c     common/ label2/ frame, kind
c     common/ label3/ cBran, polax1, polax2
c     common/ label4/ evnodd, symdes
c     common/ label5/ lngbar
c     space for calculated lines information
c     common/ lines / lqn, lqn2, intens, thefre
c     parameters determining line intensity
c     common/ linint/ temp1, temp2, weight, nuspsw
c     origin of vibronic band
c     common/ origin/ nuzero
c     output manipulating parameters
c     common/ outman/ divis, Jmxout, normf
c     ground and excited state rotational constants
c     common/ rotcon/ rotcog, rotcoe, idelta
c     vector offset pointer
c     common/ vpoint/ ivpt


      ARNIROT_LAUNCH( "Launching arnical." )

c     explicit initialization
      do k=1, 2 * npar + 1
         w( k ) = 0.0
         do i = 1, 2 * npar + 1
            v( i, k ) = 0.0
         end do
      end do

c     determine the number of actually used rotational parameters separately for ground and excited state
      nparg = npar
      do i = npar, 1, -1
         if ( rotcog(i) .ne. 0 )
     *        goto 100
         nparg = nparg - 1
      end do
  100 npare = npar
      do i = npar, 1, -1
         if ( rotcoe(i) .ne. 0 )
     *        goto 200
         npare = npare - 1
      end do
  200 continue
      ARNIROT_DEBUG3( "nparg = ", nparg, "  npare = ", npare )
      if( ( nparg .eq. 0 ) .or. ( npare .eq. 0 ) ) then
         if( shorti .eq. 0 ) then
            write(*,*) 'All rotational parameters are zero!'
            write(*,*) '...leaving krot-arnirot'
         end if
         return
      end if

c     convert polarization percentages into square root polarization factors
      do i = 1, 3
         if ( polori(i) .ne. 0 ) then
            isign = polori(i)/dabs(polori(i))
            polcal(i) = isign*dsqrt(dabs(polori(i))/1.d2)
         else
            polcal(i) = 0.d0
         end if
      end do

c     section to rotate the ground and excited state rigid rotor hamiltonians
      if ( swg .eq. 1 )
     *     call rotham( rraxsw, 1, rotcog, rotcoe, npar, cState, trang, tring, swang, shorti, lstars, lbars )
      if ( swe .eq. 1 )
     *     call rotham( rraxsw, 2, rotcog, rotcoe, npar, cState, trane, trine, swang, shorti, lstars, lbars )

c     transform transition moment if specified
      if ( itrvec .eq. 1 )
     *     call murot( trang, polcal, 3, shorti )

c     * set flags: real or complex diagonalization of hamiltonian *
c     ground state
      icomg = 0
      if ( rotcog(5) .ne. 0 )
     *     icomg = 1
      if ( rraxsw(4,1) .ne. 0 )
     *     icomg = 1
      if ( rraxsw(6,1) .ne. 0 )
     *     icomg = 1
c     excited state
      icome = 0
      if ( rotcoe(5) .ne. 0 )
     *     icome = 1
      if ( rraxsw(4,2) .ne. 0 )
     *     icome = 1
      if ( rraxsw(6,2) .ne. 0 )
     *     icome = 1

c     setup vector offset pointer array
      call ivecpt( ivpt, Jmax, Jmxcal )

c-----------------------------------------------------------------------------
c     ground state eigenvectors and eigenvalues
c-----------------------------------------------------------------------------
      if ( shorti .eq. 0 ) then
         write(*,'(a)') lstars
         write(*,*) 'ground state J block calculations:'
      end if

c     ground state J block qn counter
      do J = 0, Jmxcal, 1
         
c        current offset
         iept = J*J

c        actual size of the J block and the greatest submatrix
         iact  = 2*J + 1
         iactr = J/2 + 1

c        produce rotational state kets for both ground and excited states
         call ket( icqn, dmeval, J )

c        ground state asymmetric rotor

c        calculate Wang transformation matrix
         call wang( wangtr, dmham, iact )

c        zero the hamiltonian and eigenvector work spaces
         call matzer( hamr, dmham, iact )
         call matzer( hami, dmham, iact )
         call matzer( vecr, dmham, iact )
         call matzer( veci, dmham, iact )
         call matzer( aem, dmfwng, iactr )
         call matzer( aom, dmfwng, iactr )
         call matzer( aep, dmfwng, iactr )
         call matzer( aop, dmfwng, iactr )
         call matzer( vem, dmfwng, iactr )
         call matzer( vom, dmfwng, iactr )
         call matzer( vep, dmfwng, iactr )
         call matzer( vop, dmfwng, iactr )
c        fill ground state hamiltonian
         if ( swg .eq. 1 ) then
c           axis switched rigid rotor
            call rrhfax( J, rraxsw(1,1), rraxsw(2,1), rraxsw(3,1), hamr, dmham )
c           off-diagonal axis switching terms
            if ( rraxsw(4,1) .ne. 0 ) call jzjy( hami, dmham, J, rraxsw(4,1) )
            if ( rraxsw(5,1) .ne. 0 ) call jzjx( hamr, dmham, J, rraxsw(5,1) )
            if ( rraxsw(6,1) .ne. 0 ) call jyjx( hami, dmham, J, rraxsw(6,1) )
         else
c           rigid rotor terms not axis switched
            call watson( J, rotcog, npar, hamr, dmham )
         end if

c        linear angular momentum terms: assumed to be in axis switched frame
c        - real
         if ( rotcog(4) .ne. 0 ) call jx( hamr, dmham, J, rotcog(4) )
         if ( rotcog(6) .ne. 0 ) call jz( hamr, dmham, J, rotcog(6) )
c        - imaginary
         if ( rotcog(5) .ne. 0 ) call jy( hami, dmham, J, rotcog(5) )

#ifdef DEBUG_VERBOSE
         if ( J .le. Jmxout ) then
c           output real hamiltonian matrix for J block
            call wriham( hamr, dmham, J, kind(0), frame(swg), lngbar, lstars )
c           output imaginary hamiltonian matrix for J block
            if ( icomg .eq. 1 ) call wriham( hami, dmham, J, kind(1), frame(swg), lngbar, lstars )
         end if
#endif

c        diagonalize the hamiltonian matrix: check for imaginary hamiltonian terms
         if ( icomg .eq. 0 ) then
c           transform real hamiltonian to Wang basis
            call trwang( wangtr, hamr, wanghr, dmham, iact )
c !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c           check if the second condition is really necessary!
            if ( ( swg .eq. 1 ) .or. ( nparg .gt. 3 ) ) then
               call rotdi( wanghr, iact, dmham, ener, vecr, e, diaalg, shorti )
            else
               call rotdip( J, iact, dmham, dmfwng, wanghr, aem, aom, aep, aop,
     *                      ener, dem, dom, dep, dop, vecr, vem, vom, vep, vop,
     *                      er, diaalg, shorti )
            end if
c !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c           transform real eigenvectors back to ascending K basis from Wang basis
            call bkwang( wangtr, vecr, dmham, iact )
c           check vector matrix for any incorrect sorting
c !!!!!!!!! this procedure should be redundant after factorization of the Wang matrix
c           if ( swg .eq. 0 ) call chevec( vecr, ener, tmpvec, dmham, iact )
         else
c           transform real hamiltonian to Wang basis
            call trwang( wangtr, hamr, wanghr, dmham, iact )
c           transform imaginary hamiltonian to Wang basis
            call trwang( wangtr, hami, wanghi, dmham, iact )
            call diagc ( wanghr, wanghi, hamb, ener, enei, eneb, vecr, veci, vecb, J, dmham, eb, diaalg, shorti )
c           transform real eigenvectors back to ascending K basis from Wang basis
            call bkwang( wangtr, vecr, dmham, iact )
c           transform imaginary eigenvectors back to ascending K basis from Wang basis
            call bkwang( wangtr, veci, dmham, iact )
         end if

c        equate eigenvectors from diagonalization routine as passed back in vec matrix
         call vecsto( evecgr, dmevec, vecr, dmham, ivpt, Jmax, J )
         call vecsto( evecgi, dmevec, veci, dmham, ivpt, Jmax, J )

c        equate eigenvalues from diagonalization routine as passed back in ener array
         call ensto( evalg, dmeval, ener, dmham, iept, iact )

c        calculate rr derivatives if user wants to fit ground state parameters
         if ( fitges .eq. 1 ) then
c           Jx**2, Jy**2 and Jz**2 derivatives
            call rrde( J, derivg, dmeval, npar, dmham, iact, hamr, vecr, veci, icomg )
c           JyJx+JxJy, JzJx+JxJz and JzJy+JyJz derivatives if axis switched
            if ( swg .eq. 1 ) then
               call jyjxde( J, derivg, dmeval, npar, dmham, iact, hamr, vecr, veci )
               call jzjxde( J, derivg, dmeval, npar, dmham, iact, hamr, vecr, veci )
               call jzjyde( J, derivg, dmeval, npar, dmham, iact, hamr, vecr, veci )
c              rotate rr derivatives back to unrotated frame
               call rotder( derivg, dmeval, npar, trang, tring, iact, iept )
            end if
            if ( rotcog(4) .ne. 0 ) call jxde( J, derivg, dmeval, npar, dmham, iact, hamr, vecr, veci )
            if ( rotcog(5) .ne. 0 ) call jyde( J, derivg, dmeval, npar, dmham, iact, hamr, vecr, veci )
            if ( rotcog(6) .ne. 0 ) call jzde( J, derivg, dmeval, npar, dmham, iact, hamr, vecr, veci )
         end if

#ifdef DEBUG_VERBOSE
         if ( J .le. Jmxout ) then
c           output the real eigenvectors for J block
            call wrivec( vecr, dmham, J, icqn, dmeval, kind(0), lngbar, ldash )
c           output the imaginary eigenvectors for J block
            if ( icomg .eq. 1 ) call wrivec( veci, dmham, J, icqn, dmeval, kind(1), lngbar, ldash )
c           output exact eigenvalues
            call eigval( ener, dmham, dmeval, J, icqn, ldash )
            if ( fitges .eq. 1 ) then
c              calculate energies with derivative approximation
               call derapp( ened, dmham, iact, rotcog, npar, derivg, icqn, dmeval, J, lbars, ldash )
c              output the derivatives
               call derout( J, derivg, dmeval, npar, icqn, ldash )
            end if
         end if
#endif
      end do

c-----------------------------------------------------------------------------
c     excited state eigenvectors and eigenvalues
c-----------------------------------------------------------------------------
      if ( shorti .eq. 0 ) then
         write(*,'(a)') lstars
         write(*,*) 'excited state J block calculations:'
      end if

c     excited state J block qn counter
      do J = 0, Jmxcal, 1
         
c        current offset
         iept = J*J

c        actual size of the J block and the greatest submatrix
         iact  = 2*J + 1
         iactr = J/2 + 1

c        excited state asymmetric rotor

c        calculate the Wang transformation matrix and its inverse
         call wang( wangtr, dmham, iact )

c        zero the hamiltonian and eigenvector work spaces
         call matzer( hamr, dmham, iact )
         call matzer( hami, dmham, iact )
         call matzer( vecr, dmham, iact )
         call matzer( veci, dmham, iact )
         call matzer( aem, dmfwng, iactr )
         call matzer( aom, dmfwng, iactr )
         call matzer( aep, dmfwng, iactr )
         call matzer( aop, dmfwng, iactr )
         call matzer( vem, dmfwng, iactr )
         call matzer( vom, dmfwng, iactr )
         call matzer( vep, dmfwng, iactr )
         call matzer( vop, dmfwng, iactr )
c        fill excited state hamiltonian
         if ( swe .eq. 1 ) then
c           axis switched rigid rotor
            call rrhfax( J, rraxsw(1,2), rraxsw(2,2), rraxsw(3,2), hamr, dmham )
c           off-diagonal axis switching terms
            if ( rraxsw(4,2) .ne. 0 ) call jzjy( hami, dmham, J, rraxsw(4,2) )
            if ( rraxsw(5,2) .ne. 0 ) call jzjx( hamr, dmham, J, rraxsw(5,2) )
            if ( rraxsw(6,2) .ne. 0 ) call jyjx( hami, dmham, J, rraxsw(6,2) )
         else
c           rigid rotor terms not axis switched
            call watson( J, rotcoe, npar, hamr, dmham )
         end if

c        linear angular momentum terms: assumed to be in axis switched frame
c        - real
         if ( rotcoe(4) .ne. 0 ) call jx( hamr, dmham, J, rotcoe(4) )
         if ( rotcoe(6) .ne. 0 ) call jz( hamr, dmham, J, rotcoe(6) )
c        - imaginary
         if ( rotcoe(5) .ne. 0 ) call jy( hami, dmham, J, rotcoe(5) )

#ifdef DEBUG_VERBOSE
         if ( J .le. Jmxout ) then
c           output real hamiltonian matrix for J block
            call wriham( hamr, dmham, J, kind(0), frame(swe), lngbar, lstars )
c           output imaginary hamiltonian matrix for J block
            if ( icome .eq. 1 ) call wriham( hami, dmham, J, kind(1), frame(swe), lngbar, lstars )
         end if
#endif

c        diagonalize the hamiltonian matrix: check for imaginary hamiltonian terms
         if ( icome .eq. 0 ) then
c           transform real hamiltonian to Wang basis
            call trwang( wangtr, hamr, wanghr, dmham, iact )
c !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c           check if the second condition is really necessary!
            if ( ( swe .eq. 1 ) .or. ( npare .gt. 3 ) ) then
               call rotdi( wanghr, iact, dmham, ener, vecr, e, diaalg, shorti )
            else
               call rotdip( J, iact, dmham, dmfwng, wanghr, aem, aom, aep, aop,
     *                      ener, dem, dom, dep, dop, vecr, vem, vom, vep, vop,
     *                      er, diaalg, shorti )
            end if
c !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c           transform real eigenvectors back to ascending K basis from Wang basis
            call bkwang( wangtr, vecr, dmham, iact )
c           check vector matrix for any incorrect sorting
c !!!!!!!!! this procedure should be redundant after factorization of the Wang matrix
c           if ( swe .eq. 0 ) call chevec( vecr, ener, tmpvec, dmham, iact )
         else
c           transform real hamiltonian to Wang basis
            call trwang( wangtr, hamr, wanghr, dmham, iact )
c           transform imaginary hamiltonian to Wang basis
            call trwang( wangtr, hami, wanghi, dmham, iact )
            call diagc ( wanghr, wanghi, hamb, ener, enei, eneb, vecr, veci, vecb, J, dmham, eb, diaalg, shorti )
c           transform real eigenvectors back to ascending K basis from Wang basis
            call bkwang( wangtr, vecr, dmham, iact )
c           transform imaginary eigenvectors back to ascending K basis from Wang basis
            call bkwang( wangtr, veci, dmham, iact )
         end if

c        equate eigenvectors from diagonalization routine as passed back in vec matrix
         call vecsto( evecer, dmevec, vecr, dmham, ivpt, Jmax, J )
         call vecsto( evecei, dmevec, veci, dmham, ivpt, Jmax, J )

c        equate eigenvalues from diagonalization routine as passed back in ener array
         call ensto( evale, dmeval, ener, dmham, iept, iact )

c        calculate rr derivatives if user wants to fit excited state parameters
         if ( fitges .eq. 1 ) then
c           Jx**2, Jy**2 and Jz**2 derivatives
            call rrde( J, derive, dmeval, npar, dmham, iact, hamr, vecr, veci, icome )
c           JyJx+JxJy, JzJx+JxJz and JzJy+JyJz derivatives if axis switched
            if ( swe .eq. 1 ) then
               call jyjxde( J, derive, dmeval, npar, dmham, iact, hamr, vecr, veci )
               call jzjxde( J, derive, dmeval, npar, dmham, iact, hamr, vecr, veci )
               call jzjyde( J, derive, dmeval, npar, dmham, iact, hamr, vecr, veci )
c              rotate rr derivatives back to unrotated frame
               call rotder( derive, dmeval, npar, trane, trine, iact, iept )
            end if
            if ( rotcoe(4) .ne. 0 ) call jxde( J, derive, dmeval, npar, dmham, iact, hamr, vecr, veci )
            if ( rotcoe(5) .ne. 0 ) call jyde( J, derive, dmeval, npar, dmham, iact, hamr, vecr, veci )
            if ( rotcoe(6) .ne. 0 ) call jzde( J, derive, dmeval, npar, dmham, iact, hamr, vecr, veci )
         end if

#ifdef DEBUG_VERBOSE
         if ( J .le. Jmxout ) then
c           output the real eigenvectors for J block
            call wrivec( vecr, dmham, J, icqn, dmeval, kind(0), lngbar, ldash )
c           output the imaginary eigenvectors for J block
            if ( icome .eq. 1 ) call wrivec( veci, dmham, J, icqn, dmeval, kind(1), lngbar, ldash )
c           output exact eigenvalues
            call eigval( ener, dmham, dmeval, J, icqn, ldash )
            if ( fitges .eq. 1 ) then
c              calculate energies with derivative approximation
               call derapp( ened, dmham, iact, rotcoe, npar, derive, icqn, dmeval, J, lbars, ldash )
c              output the derivatives
               call derout( J, derive, dmeval, npar, icqn, ldash )
            end if
         end if
#endif
      end do

c-----------------------------------------------------------------------------
c     line frequencies and intensities
c-----------------------------------------------------------------------------
      if ( shorti .eq. 0 ) then
         write(*,'(a)') lstars
         write(*,*) 'evaluating line intensities'
      end if

      cutor = cutint

c     start counter for number of calculated lines
  600 ntheli = 0

#ifdef DEBUG_VERBOSE
      open(30, file = 'debug.P', status = 'unknown')
      open(40, file = 'debug.Q', status = 'unknown')
      open(50, file = 'debug.R', status = 'unknown')
#endif

c     calculate all P, Q and R branch transitions J by J
      call pqr( Jmax, dmeval, dmevec, maxnli, npar,
     *          ntheli, icqn, lioflw, icomg+icome,
     *          derivg, derive, derlg, derle,
     *          evalg, evale,
     *          evecgr, evecgi, evecer, evecei,
     *          Jmxcal, dKmax, cutint,
     *          polcal,
     *          kind,
     *          cBran, polax1,
     *          lngbar, lbars, ldash,
     *          lqn, lqn2, intens, thefre,
     *          temp1, temp2, weight, nuspsw,
     *          nuzero,
     *          Jmxout,
     *          ivpt )

#ifdef DEBUG_VERBOSE
      close(30)
      close(40)
      close(50)
#endif
      
      if ( lioflw .ne. 0 ) then
c        increase cutint in a convenient way
         cutint = dmin1(cutint*2.d0, cutint + 2.d0)
         if ( shorti .eq. 0 ) then
            write(*,'(a)') lstars
            write(*,'(a,i3)') 'ERROR  : overflow in number of lines during calculation for J =', J
            write(*,'(9x,a,f9.6)') 'recalculating with new value of cutint:', cutint
         end if
         goto 600
      end if

      if ( ( cutint .ne. cutor ) .and. ( shorti .eq. 0 ) ) then
         write(*,'(a)') lstars
         write(*,'(a,f9.6)') 'WARNING: actual value for cutint is ', cutint
         write(*,'(a)') 'You should adopt this value into your input.'
      end if

c-----------------------------------------------------------------------------
c     statistical least squares fitting section
c-----------------------------------------------------------------------------
      if ( fitges .eq. 1 ) then
c        we want the fit, fitges gets back to 1 when the fit is done
         fitges = 0
         if ( shorti .eq. 0 ) then
            write(*,'(a)') lstars
            write(*,*) 'least squares analysis:'
            write(*,'(a)') lstars
         end if

c        restructure theoretical lines set in terms of packed quantum numbers
         call qnpac1( pqnth, lqn, maxnli, ntheli )

c        sort derivative and line information by packed qn
         call linord( maxnli, npar, ntheli, nparg, npare,
     *                iop, derlg, derle, pqnth,
     *                lqn, lqn2, intens, thefre )

c        read an experimental lines file and establish correspondence between experimental and theoretical line frequencies
         call expthe( maxnli, ntheli, nexpli, pqnth, asfile,
     *                shorti,
     *                lqn, thefre, expfre )

c        number of theoretical data sets matching to experimental lines
         ndata = 0
         if( nexpli .le. 0 ) then
            write(*,*) "No experimental lines !\n...leaving the fitting section"
            return
         end if

c        sum of squared deviations (observed minus calculated line positions)
         sdsum = 0.d0            
c        establish index to theoretical lines with experimental assignments
         do i = 1, ntheli, 1
            if ( expfre(i) .ne. 0 ) then
               ndata = ndata + 1
               sdsum = sdsum + (expfre(i)*divis - thefre(i))**2
c              index the theoretical lines which are matched to experimental ones
               ipttem(ndata) = i
            end if
         end do
         if ( shorti .eq. 0 ) write(*,'(a,i6)') 'number of assignments matching calc. lines   :', ndata

c        output lines and leave this program part if zero data
         if ( ndata .eq. 0 ) then
            if ( shorti .eq. 0 ) then
               write(*,'(/,a)') 'There are no assigned lines matching calculated transitions!'
               write(*,'(a)') '...leaving the fitting section'
            end if
            return
         end if

c        load the indexed derivatives into design matrix for least squares analysis
         call funcsl( funcs, maxnli, npar, ndata, ipttem, derlg, derle )

c        if fitting delta constants ...
         if ( idelta .eq. 1 ) then
c           ... adapt array funcs to delta derivatives
            call delder( funcs, npar, maxnli, ndata )
c           ... convert the rotational constants to delta constants
            call delabs( rotcog, rotcoe, npar, 1 )
         end if

c        loop to load unit derivative for nuzero and experimental frequency vector
         do i = 1, ndata, 1
c           y = experimental frequency adjusted for ASYROT factor
            y(i) = expfre(ipttem(i))*divis
c           nuzero derivative is always one
            funcs(i,2*npar+1) = 1.d0
         end do

c        counter of active parameters: all 2*npar+1 are active at beginning
         nparft = 2*npar+1
         if ( shorti .eq. 0 ) write(*,'(/,a,i6)') 'maximum number of parameters to be fit       :', nparft
c        sum constant contributions of invariant parameters to y
         icount = 0

c        loop through ground state contributions
         do k = 1, npar, 1
            if ( ifit(k) .eq. 0 ) then
               nparft = nparft - 1
               do i = 1, ndata, 1
                  y(i) = y(i) - funcs(i,k)*rotcog(k)
c                 zero that column of funcs
                  funcs(i,k) = 0.d0
               end do
            else
               icount = icount + 1
               ifitpt(icount) = k
            end if
         end do
c        loop through excited state contributions
         do k = 1, npar, 1
            if ( ifit(k+npar) .eq. 0 ) then
               nparft = nparft - 1
               do i = 1, ndata, 1
                  y(i) = y(i) - funcs(i,k+npar)*rotcoe(k)
c                 zero that column of funcs
                  funcs(i,k+npar) = 0.d0
               end do
            else
               icount = icount + 1
               ifitpt(icount) = k + npar
            end if
         end do
c        nuzero contribution
         k = 2*npar + 1
         if ( ifit(k) .eq. 0 ) then
            nparft = nparft - 1
            do i = 1, ndata, 1
               y(i) = y(i) - nuzero
c              zero that column of funcs
               funcs(i,k) = 0.d0
            end do
         else
            icount = icount + 1
            ifitpt(icount) = k
         end if

         if ( shorti .eq. 0 ) then
            write(*,'(a,15x,a,i6)') 'number of parameters being fit',':', nparft
            write(*,'(/,23(i4))') (ifitpt(i), i = 1, nparft, 1)
         end if

c        if no parameters are active just output lines and finish
c        reset fitting flags to prevent program from breaking down at subroutine idef (utility.F)
         if ( ( nparft .eq. 0 ) .or. ( nparft .gt. ndata ) ) then
            if ( shorti .eq. 0 ) then
               if ( nparft .eq. 0 ) then
                  write(*,'(/,a)') 'WARNING: All empirical parameters are inactive!'
               else
                  write(*,'(/,a,i2,a,i2,a)') 'ERROR: ARNIROT cannot solve for ',nparft,' parameters with only ',ndata,' sets of data!'
               end if
               write(*,'(a)') '...leaving the fitting section'
            end if
            return
         end if
         if ( shorti .eq. 0 ) write(*,'(a)') lbars

c        rearrange matrix FUNCS so that all nonzero columns are in the first NPARFT positions
c        save FUNCS to a temporary array FUNCST
         do k = 1, 2*npar+1, 1
            do i = 1, ndata, 1
               funcst(i,k) = funcs(i,k)
            end do
         end do
c        copy into correct position
         do k = 1, nparft, 1
            do i = 1, ndata, 1
               funcs(i,k) = funcst(i,ifitpt(k))
            end do
         end do

c        statistical analysis routine to empirically fit the rotational parameters
         call svdfit( y, sigma, ndata, a, nparft, u, v, w, maxnli, 2*npar+1, chisq, funcs, shorti )

c        calculate and output the covariance marix
         call svdvar( v, 2*npar+1, 2*npar+1, w, cvm, 2*npar+1 )

         if ( shorti .eq. 0 ) then
            write(*,'(/,a,/)') 'covariance matrix (part 1 of 4):'
            do i = 1, 2*npar+1, 1
               write(*,'(6(d13.4))') (cvm(i,k), k =  1, 6, 1)
            end do
            write(*,'(/,a,/)') 'covariance matrix (part 2 of 4):'
            do i = 1, 2*npar+1, 1
               write(*,'(6(d13.4))') (cvm(i,k), k =  7,12, 1)
            end do
            write(*,'(/,a,/)') 'covariance matrix (part 3 of 4):'
            do i = 1, 2*npar+1, 1
               write(*,'(6(d13.4))') (cvm(i,k), k = 13,18, 1)
            end do
            write(*,'(/,a,/)') 'covariance matrix (part 4 of 4):'
            do i = 1, 2*npar+1, 1
               write(*,'(6(d13.4))') (cvm(i,k), k = 19,23, 1)
            end do
         end if

c        load the correct covariance term into an error vector
         do i = 1, nparft, 1
            err(ifitpt(i)) = dsqrt(cvm(i,i))
         end do

c        rearrange A (fit parameters) back to initial order
         do i = 1, 2*npar+1, 1
            atmp(i) = a(i)
         end do
c        load original constants back in
         do i = 1, npar, 1
            a(i)      = rotcog(i)
            a(i+npar) = rotcoe(i)
         end do
         a(2*npar+1) = nuzero
c        load in the A constants that have been changed
         do i = 1, nparft, 1
            a(ifitpt(i)) = atmp(i)
         end do

         if ( shorti .eq. 0 ) then
c           output least squares analysis summary
            write(*,'(a)') lstars
            write(*,*) 'least squares analysis summary:'
            write(*,'(a)') lstars
c           output initial and final values of constants with standard deviations
            write(*,'(a,/)') 'ground state constants  [MHz]:'
            do i = 1, npar, 1
               write(*,'(a12,''old:'',f14.6,2x,''new:'',f14.6,''+-'',f14.6,2x,''new-old:'',f14.6)')
     c         cRotCg(i), rotcog(i), a(i), err(i), a(i) - rotcog(i)
            end do
            write(*,'(a)') lbars
            write(*,'(a,/)') 'excited state constants [MHz]:'
            do i = 1, npar, 1
               if (idelta.eq.0) then
                  write(*,'(a12,''old:'',f14.6,2x,''new:'',f14.6,''+-'',f14.6,2x,''new-old:'',f14.6)')
     c            cRotCe(i), rotcoe(i), a(i+npar), err(i+npar), a(i+npar) - rotcoe(i)
               else
                  write(*,'(a12,''old:'',f14.6,2x,''new:'',f14.6,''+-'',f14.6,2x,''new-old:'',f14.6)')
     c            cDRotC(i), rotcoe(i), a(i+npar), err(i+npar), a(i+npar) - rotcoe(i)
               end if
            end do
c           output nuzero data
            write(*,'(a)') lbars
            write(*,'(/,''nuzero      old:'',f15.3,'' MHz'',/,12x,''new:'',f15.3,'' +- '',f9.3,'' MHz'',/,8x,''new-old:'',f15.3,'' MHz'')')
     c      nuzero, a(2*npar+1), err(2*npar+1), a(2*npar+1) - nuzero

            write(*,'(a)') lbars
c           uncertainty assumed for each line assignment
            write(*,'(''assumed uncertainty for each line assignment :'',f10.3,'' MHz'')') sigma
c           number of calculated lines
            write(*,'(''number of generated theoretical lines        :'',i6)') ntheli
c           number of assigned lines
            write(*,'(''number of assigned experimental lines        :'',i6)') nexpli
c           number of experimental lines which have been matched to theoretical lines
            write(*,'(''number of exp. lines matched to theor. lines :'',i6)') ndata
c           average observed minus calculated line positions
            write(*,'(''average observed-calculated line position    :'',f10.3,'' MHz'')') dsqrt(sdsum/ndata)
         end if

c        equate old rotational constants to new improved least squares constants
         do i = 1, npar, 1
            rotcog(i) = a(i)
            rotcoe(i) = a(i+npar)
         end do
         nuzero = a(2*npar+1)
c        calculate new inertial defect and uncertainties
         if ( shorti .eq. 0 ) then
            write(*,'(a)') lbars
            call ideferr( rotcog, rotcoe, npar, err, idelta )
         end if
         fitges = 1
      end if

      return
      end
