// This file is part of krot,
// a program for the simulation, assignment and fit of HRLIF spectra.
//
// Copyright (C) 1998 Jochen K"upper
//
// $Id: selectSet.cc,v 1.3 1998/09/15 10:42:55 jochen Exp $


#include "selectSet.h"

#include "dataset.h"
#include "krotData.h"
#include "krot.h"

#include <iostream.h>
#include <string>

#include <qdialog.h>
#include <qlayout.h> 
#include <qpushbt.h> 

#include <kapp.h>
#include <kbuttonbox.h>
#include <klocale.h>
#include <kmsgbox.h> 




DialogSelectSet::DialogSelectSet( KRotData *krotdata, QWidget *parent, const char *name )
    : QDialog(parent, name, true),
      krotData( krotdata )
{
    KROT_LAUNCH( "Launching DialogSelectSet constructor" );
    setCaption( i18n("KRot [Select set]") );
    QBoxLayout *layout = new QVBoxLayout( this );
    // *** listbox ***
    list = new QListBox(this, "Listbox",
			WStyle_Customize | WStyle_DialogBorder | WStyle_SysMenu | WStyle_Title );
    // insert items for all sets
    list->setAutoUpdate( false );
    mapSpectrum::const_iterator spec = krotData->spectra().begin();
    while( spec != krotData->spectra().end() ) {
	list->inSort( spec->first.data() );
	spec++;
    }
    mapSimulation::const_iterator sims = krotData->simulations().begin();
    while( sims != krotData->simulations().end() ) {
	list->inSort( sims->first.data() );
	sims++;
    }
    list->setAutoUpdate( true );
    layout->addWidget( list );
    // *** buttons ***
    KButtonBox *bbox = new KButtonBox(this);
    bbox->addStretch( 5 );
    QPushButton *ok = bbox->addButton( i18n( "Select" ) );
    if( list->count() == 0 )
	ok->setEnabled( false );
    bbox->addStretch( 10 );
    QPushButton *cancel = bbox->addButton( i18n( "Cancel" ) );
    bbox->addStretch( 10 );
    QPushButton *help   = bbox->addButton( i18n( "Help" ) );
    bbox->addStretch( 5 );
    bbox->layout();
    bbox->setFixedHeight( bbox->sizeHint().height() );

    layout->addWidget( bbox );
    layout->activate();
    // connect signals
    connect( list, SIGNAL(selected(const char *)), this, SLOT(select(const char *)) );
    connect( ok, SIGNAL(clicked()), this, SLOT( accept() ) );
    connect( cancel, SIGNAL(clicked()), this, SLOT(reject()) );
    connect( help, SIGNAL(clicked()), this, SLOT(help()) );
    setGeometry( x(), y(), bbox->sizeHint().width(), height() );
    return;
}



const QString DialogSelectSet::getSelection()
{
    KROT_LAUNCH( "Launching DialogSelectSet::getSelection" );
    if( DialogSelectSet::Accepted == exec() ) {
	return list->text( list->currentItem() );
    } else
	return "";
}



void DialogSelectSet::select()
{
    KROT_LAUNCH( "Launching DialogSelectSet::select()" );
    accept();
    return;
}



void DialogSelectSet::select(const char *)
{
    KROT_LAUNCH( "Launching DialogSelectSet::select(const char *)" );
    accept();
    return;
}



void DialogSelectSet::help()
{
    kapp->invokeHTMLHelp("","");
    return;
}
